# Makefile.M for extra/src/c-client directory
# $Id: Makefile.M,v 1.7 2000/12/05 08:22:53 gregnoel Exp $

# This is the "real" Makefile for the c-client directory.  It kludges
# around the Makefile provided with the library, coercing it to work
# with the Mahogany build system.

include ../../../makeopts

# we depend upon the fact that the last file created by c-client's
# setup is $(SET_UP), so if it is present, we're ready to go.
SET_UP := osdep.c

# c-client cannot use a parallel make.  This may be excessive paranoia,
# but we go to a lot of trouble to avoid ever invoking it in parallel.
# strip off any -j arguments to environmental MAKE
MAKE := $(shell echo $(MAKE) | sed "1,1 s/-j[ 0-9]*//g")
# pick a single target in case we are parallel
TARGET := $(shell if test -f $(SET_UP); then echo c-client.a; else echo $(SET_UP); fi)

all:	$(TARGET)
	@true

# only tripped if using a separate build directory
drivers: .src/drivers
	ln -sf .src/* .

# initial make call, set up for our OS type
$(SET_UP): drivers
	@$(MAKE) -j1 $(CCLIENTOSTYPE) --no-print-directory \
		MAKE="$(MAKE) -j1" CC="$(CC)" \
		EXTRAAUTHENTICATORS="$(CCLIENT_EXTRAAUTHENTICATORS)" \
		EXTRACFLAGS="$(CCLIENT_EXTRACFLAGS)"

# subsequent make call, only call c-client's Makefile if it looks
# like there's something to do
c-client.a: $(wildcard *.c) $(wildcard *.h)
	@$(MAKE) -j1 --no-print-directory MAKE="$(MAKE) -j1"

clean:
	@$(MAKE) -j1 --no-print-directory clean 
	$(RM) X `find * -type l ! -name Makefile -print`

.PRECIOUS: $(SET_UP)
.PHONY: all clean
